/**
* \file single.cpp
*/
#include "IPlayableState.hpp"
#include <cmath>
#include <string>
#include <irrlicht.h>
#include "../entities/player.hpp"
#include "../entities/IPlayerAnimator.hpp"
#include "../config.hpp"
#include "../game_engine.hpp"
//#include "../states/CMenuState.hpp"
#include "../states/COptionsState.hpp"
#include "../states/CIngamePauseState.hpp"
#include "../states/CScoreState.hpp"
#include "../states/CErrorState.hpp"
#include "../IdList.hpp"
#include "../fx/CParticleManager.hpp"
#include "../utils/CBitSet.hpp"
//#include "../entities/IMultiModeTransformer.hpp"
#include <fusion/scripting/CLuaThread.hpp>
#include <fusion/scripting/CLuaVM.hpp>
#include <fusion/maths/macros.hpp>
#include <fusion/core/string.hpp>
#include "../entities/IWalkyrie.hpp"
#include "../fleets/CEnemyFleet.hpp"
#include "../fleets/CAlliedFleet.hpp"

#include <mirrpp/CScreenQuadCB.hpp>
#include <mirrpp/CAddTransparencyCB.hpp>

const int _OSCILLATION_NUMBER = 3;


using namespace irr;
using namespace video;
using namespace core;
using namespace scene;


//DECLARE_GAME_SINGLETON( IPlayableState )





// Passer le nom du script en parametre ?

//! Constructor
IPlayableState::IPlayableState(const std::string& name) :
IGameState(name,false,false,true,false,false)
//, Player(0)
//         *CGameEngine::instance().getPlayer() )
//, Level(0)
//, Scene(0)
, _blinksCounter(0)
, _alliedFleet(0)
, _enemyFleet(0)
//, Progression(0)
, CurrentPhase(Phase::Playing)  //! TODO
, _rttHud(0)
, _rttScene(0)
, _lifebar(0)
, _lifePictogram(0)
, _lifebarMat(-1)
, _cannonPictogram(0)
, _heatbar(0)
//new CScreenQuadWithShader()
, _lifebarCB( new CLifebarCB() )
, _customQuad( new CScreenQuadWithShader() )
, _cachedFleetNode(0)
, _activeFleetNode(0)
, ShowDebugData(false)
//, _playerMapping( &CGameEngine::instance().GameConfig.MappingPlayer1 )
{

//    ADD_REQUIRED( _rttHud, "RTT HUD");
//    ADD_REQUIRED( _rttScene, "RTT Scene");


//    Player->set ;
};

//! Destructor
IPlayableState::~IPlayableState()
{
    //_world->remove();
//    delete _script;

    delete _lifebarCB;
    delete _customQuad;
}



bool
IPlayableState::load(CGameEngine* engine,irr::scene::ISceneManager* smgr,irr::video::ITexture* screenshot)
{
    if (! IGameState::load(engine,smgr,screenshot) ) return false;

    ///////////////////////////////////////////:
    ////    setup player
    ///////////////////////////////////////////
    IWalkyrie* walkyrie = engine->getWalkyrie(NWalkyrieModel::YF19) ;
    FUS_ASSERT( walkyrie != nullptr);


    getPlayer(0).setWalkyrie( walkyrie );
//    walkyrie->allowMode( EWalkyrieMode::Battloid );
//    walkyrie->allowMode( EWalkyrieMode::Fighter );
//    walkyrie->allowMode( EWalkyrieMode::Gerwalk );

    //! Create fleets
    // TODO lui passer les gestionnaires d'explosions, et de son
    //
    //engine().getMusicManager()
    //
    irr::scene::ISceneNode *_cachedFleetNode = smgr->addEmptySceneNode(0);
    irr::scene::ISceneNode *_activeFleetNode = smgr->addEmptySceneNode(0);



    _enemyFleet = new CEnemyFleet(
//                        engine().getSoundManager(),
//                        engine().getParticleManager()
                        this
                        );



    _alliedFleet = new CAlliedFleet(
                            this
                            );

    //! add

    getPlayer(0).setFleet( getAlliedFleet() );

//    int bandWidth = conf::_SIDE_BAND_PERCENTAGE*dim.Width;
//    area.UpperLeftCorner = position2di(bandWidth,0);
//    area.LowerRightCorner  = position2di(dim.Width-bandWidth,dim.Height);
    core::dimension2du rttSceneDims ( this->getSceneDimensions() );
    //driver()->getScreenSize() - dimension2du( 2* conf::SIDEBAND_WIDTH_ABSOLUTE,0) );


        //_WINDOW_MARGIN_LEFT
//    _rttHud = driver()->addRenderTargetTexture(
//
//                                                driver()->getScreenSize()
//                                                , "rtt HUD"
////                                                , ECF_R32F
////engine().getRRTformat() (le chercher dans les options)
//
//                                                );


    _rttScene = driver()->addRenderTargetTexture(
                                                rttSceneDims
                                                , "rtt scene"
//                                                , ECF_R32F
                                                );

    //! Load HUD textures
    _lifebar            = driver()->getTexture( "images/hud/lifebar.png" );
    _lifePictogram      = driver()->getTexture( "images/hud/lifePictogram.png" );
    _heatbar            = driver()->getTexture( "images/hud/heatbar.png" );
    _cannonPictogram    = driver()->getTexture( "images/hud/cannonPictogram.png" );

    //! load shader for
    IGPUProgrammingServices* gpu = driver()->getGPUProgrammingServices();

    const E_VERTEX_SHADER_TYPE vertexProfile = pp()->getVertexProfile();
    const E_PIXEL_SHADER_TYPE pixelProfile = pp()->getPixelProfile();


    _lifebarCB->UpperLimit = 0.8f;

    //! setup material
    _lifebarMat =   gpu->addHighLevelShaderMaterialFromFiles(
                    "shaders/2d.vert.glsl", "main",vertexProfile,
                    "shaders/lifebar.frag.glsl", "main", pixelProfile
                    , _lifebarCB
                    , video::EMT_TRANSPARENT_ALPHA_CHANNEL
                    );



//    _lifebarShader  = gpu->addHighLevelShaderMaterialFromFiles(
//                    "shaders/motion_blur.vert", "main", pp()->getVertexProfile(),
//                    "shaders/lifebar.glsl.frag", "main", pp()->getPixelProfile(),
//                    0, //! Callback
//                    video::EMT_SOLID);


    // TODO creer _cachedFleetNode et _activeFleet


//    IMultiModeTransformer<int> hello;

    return true;
}



scene::IAnimatedMeshSceneNode*
IPlayableState::addAnimatedNode()
{
    //
//        scene::IAnimatedMeshSceneNode*    node = 0;
    return scene()->addAnimatedMeshSceneNode( 0,  // fleet
                                            0,   // parent
                                            -1, //id
                                            core::vector3df(0, 0, 0),   //position
                                            core::vector3df(0, 0, 0), //rotation
                                            core::vector3df(1.0f, 1.0f, 1.0f), // scale
                                            true        // add node even without mesh
                                            );
}


//
//void
//IPlayableState::updateInput()
//{
//    //! shouldn't be done from CGameEngine ?
////    engine().GameConfig.MappingPlayer1.updateCache( engine().device()->getTimer()->getTime(), _cachedInput );
//
//    engine().updateCacheMappingForPlayer1( _cachedInput );
//
//}


void
IPlayableState::setupPlayerForNewGame()
{
    //_trackedMissile(0),
//    PlayerStats.LivesLeft = conf::_DEFAULT_LIVES;
//    getPlayer(0).getWalkyrie()->forceMode( NWalkyrieModes::yf19_Fighter);
    getPlayer(0).getWalkyrie()->forceMode( IWalkyrie::FighterMode );
    getPlayer(0).setupNewLife();
}




irr::u32
IPlayableState::getElapsedTime() const {
    return engine().getElapsedTime();
}



irr::scene::IAnimatedMesh*
IPlayableState::getMesh(
                        const char* filename
                        )
{
    //createBatteryModel
    //CGameEngine* engine = CGameEngine::getInstance();
    //CGameEngine* engine = ENGINE;
    irr::scene::IAnimatedMesh* mesh = scene()->getMesh(filename);

    if(!mesh){
        _LOG_WARNING << "could not load mesh [" << filename << "]";
        return 0;

    }
    return mesh;
//
//    CModel *model = new CModel();
//
//    model->_mesh = mesh;
//    model->_bonus= bonus;
//    model->_range = range;
//    model->_life = life;
//    return model;
}




void
IPlayableState::addSimpleExplosion(const TVector3df& position)
{
//
//    ISceneNode* node = engine().getParticleManager()->createExplosion();
//
//    if( node == nullptr){
//        _LOG_WARNING << "Explosion not created";
//        return;
//    }
//    node->setPosition( position );
//    node->setParent( scene()->getRootSceneNode() );
//    node->setScale( TVector3df( 4,4,4) );
    //

}








//
//void
//IPlayableState::onResume()
//{
//    //_levelMgr.newGame();
//    //getLevelManager()._playState = CGraphicalLevel::Playing;
//    _player->setupForNewGame();
//}

//void  IPlayableState::onHold() ;



//! should be dirty

//void
//IPlayableState::renderHUD(  irr::video::ITexture* hud)
//{
//    //hud
//}






// TODO deplacer ca dans lua
//const float& playerAdv,
//, IPlayableState* _state
/*
void
IPlayableState::update(const TTimer::TimeType& elapsedTime){




//    _layer->getMaterial(0).getTextureMatrix(0).setTextureTranslate(0.f,-percievedSpeed);




    speed = _player->_walkyrie->getSpeed(SModeDescription::DefaultForward)/100.f;

    // plus on est eloigne du bas de l'ecran plus la sensation de vitesse
    float ecart = (_player->getPosition().Z- conf::_LOWER_LIMIT);
    ecart /= (conf::_UPPER_PART - conf::_LOWER_LIMIT);
    // Comme ca on a un ecart < 1

    //_INFO << "ecart " << ecart;
    speed += ecart/1000.f;// /10.f


    //_INFO << "upper progress" << _levelMgr._upperProgress/100.f;

    percievedSpeed += (speed);

    // TOOOOOOODOOOO : a retirer !!
    //BUG
    getLevelManager().updatePlayerProgress(percievedSpeed);


    //_INFO << "speed" << translation;

    // Selon l'avancement du joueur, incliner plus ou moins la camera
    this->updateCamera();


}
*/



irr::core::dimension2du
IPlayableState::getSceneDimensions() const
{
//    int bandWidth = conf::_SIDE_BAND_PERCENTAGE*dim.Width;
//    area.UpperLeftCorner = position2di(bandWidth,0);
//    area.LowerRightCorner  = position2di(dim.Width-bandWidth,dim.Height);
//rttSceneDims
    return core::dimension2du ( driver()->getScreenSize() - dimension2du( 2* conf::SIDEBAND_WIDTH_ABSOLUTE,0) );
}


std::wstring
IPlayableState::generateDebugData() const
{
    fus::TCharStringBuilder temp;


    const CPlayer& player = Player;
    const IWalkyrieMode* mode = player.getWalkyrie()->getCurrentMode();

    const auto cachedInput = engine().getCachedInput();

    temp("Player mode: ")( mode->Name )
        ("\nMay transform into: \n-Previous: ");
        if(mode->getPreviousMode())     temp( mode->getPreviousMode()->Name);
        else        temp("None");
        temp("\n-Next: ");
        if(mode->getNextMode())     temp( mode->getNextMode()->Name);
        else        temp("None");
    temp("\nMove func: ")( player.UpdatePositionFunction->Name );


    temp("\nWalkyrie name: ")( player.getWalkyrie()->Name )
        ("\nPlayer lives: ")( player.getNumberOfLivesLeft() )
        ("\nPlayer Pos: ")( player.getPosition().X )("/")( player.getPosition().Y )("/")( player.getPosition().Z )
        ("\nPlayer Rot: ")( player.getRotation().X )("/")( player.getRotation().Y )("/")( player.getRotation().Z )
        ("\nTransforming")( (player.getWalkyrie()->isTransforming()) ? L"True" : L"False" )
        ("\nOverheat: ")( (player.coolingDown()) ? L"True" : L"False")
        ("\nHeat ratio: ")( player.getCannonHeatRatio() )
//        ("\nFleet size: ")( getAlliedFleet()->getNumberOfActiveEntities() )
//        ("\nCached bullets: ")( getAlliedFleet()->getNumberOfCachedEntities(EEntityType::Bullet) )
//        ("\nCached missiles: ")( getAlliedFleet()->getNumberOfCachedEntities(EEntityType::Missile) )
//        ("\nCached static batteries: ")( getAlliedFleet()->getNumberOfCachedEntities(EEntityType::Battery) )
        ;

    temp("\nBindings state:");
            //
        const auto up = cachedInput.find(NPlayerInput::MoveForward);
        const auto down = cachedInput.find(NPlayerInput::MoveBackward);
        const auto left = cachedInput.find(NPlayerInput::MoveLeft);
        const auto right = cachedInput.find(NPlayerInput::MoveRight);
        //cachedInput[NPlayerInput::MoveForward]
    temp("\nUp: ")(  (up->second.pressed() ) )
        ("\nDown: ")(  (down->second.pressed() ) )
        ("\nLeft: ")(  (left->second.pressed() ) )
        ("\nRight: ")(  (right->second.pressed() ) );
//        ("\nLeft")

    std::wstring wtemp;

    fus::convertIntoWideString( temp.str(), wtemp );
//    text.append( wtemp.c_str() );
    return wtemp;

//    text.append( L"Player lives: ").append( getPlayer(0).getNumberOfLivesLeft() )
//        .append( L"\nTransforming: ").append( (getPlayer(0).getWalkyrie()->isTransforming()) ? L"True" : L"False")
//        .append( L"\nOverheat: ").append( (getPlayer(0).coolingDown()) ? L"True" : L"False")
//        .append( L"\nCannon heat: ").append( getPlayer(0).getCannonHeatRatio() );

}





//! Render a frame
// en gros l'ecran a une taille de 60
void
IPlayableState::Draw()
{

//    _LOG_DEBUG << "starting drawing IPlayableState";

    const irr::core::dimension2du& screenSize =  engine().getScreenSize();
//    TVector3df start , end;


    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///// If player recently died, blink the ship
    ////////////////////////////////////////////////////
//    if(_blinksCounter > 0)
//    {
//
//        //_elapsedTimes[SinceBeingKilled] += elapsedTime;
//
//        _elapsedTimes[SinceLastBlinkRevival] += elapsedTime;
//
//        if(_elapsedTimes[SinceLastBlinkRevival] > conf::_PLAYER_BLINK_LENGTH){
//
//            --_blinksCounter;
//            _elapsedTimes[SinceLastBlinkRevival] = 0;
//
//            // renders or not the node
//            if(_blinksCounter <= 0){
//                setupNewLife();
//            }
//            else {
//                _node->setVisible(! _node->isVisible() );
//            }
//        }
//    }

    //! draw scene into rtt
    driver()->setRenderTarget( _rttScene );

    // Draw scene

    scene()->drawAll();

//    if(ShowDebugData)
//    {
    renderDebugData();
//    }

    //! draw ,true,true
    driver()->setRenderTarget(ERT_FRAME_BUFFER);



    // render scene
    pp()->render( _rttScene, position2di( conf::SIDEBAND_WIDTH_ABSOLUTE, 0) );





    ////////////////////////////////////////////
    ///// Heat
    ////////////////////////////////////////////
    _customQuad->getMaterial().MaterialType = (E_MATERIAL_TYPE)_lifebarMat;
    position2di heatbarPos (  screenSize.Width - conf::SIDEBAND_WIDTH_ABSOLUTE + 10 , screenSize.Height - _heatbar->getSize().Height - 10 ) ;

    _lifebarCB->UpperLimit = getPlayer(0).getCannonHeatRatio();
    _customQuad->render( driver()
                          , heatbarPos
                          , _heatbar
                          );


    //! draw cannon nearby
    position2di cannonPictoPos (  heatbarPos - vector2di(0, _lifePictogram->getSize().Height + 5 )  ) ;
    _lifebarCB->UpperLimit = 0.f;
    _customQuad->render( driver(),
                cannonPictoPos ,
                _cannonPictogram
                );
//
//    vector2di(10,10);

//PLAYER_CANNON_MAX_HEAT-




    ////////////////////////////////////////////
    ///// Draw lifebar
    ////////////////////////////////////////////
    position2di lifebarPos ( 10, screenSize.Height - _lifebar->getSize().Height - 20 ) ;
//    driver()->draw2DImage(
    _customQuad->render( driver()
                          , lifebarPos
                          , _lifebar
                          );


    ////////////////////////////////////////////
    /////   Number of lifes left
    ////////////////////////////////////////////
    vector2di lifePictoPos ( 10, 10);
    for( int i (0); i < getPlayer(0).getNumberOfLivesLeft(); ++i)
    {
        _customQuad->render( driver()
                              , lifePictoPos
                              , _lifePictogram
                              );

        lifePictoPos += vector2di ( 0, _lifePictogram->getSize().Height + 5);
    }



    ////////////////////////////////////////////
    /////   Missiles
    ////////////////////////////////////////////
//    driver->draw2DRectangle(
//                heatColor,
//                core::recti( upperLeft + core::vector2di(40,0), core::dimension2du(10, getPlayer(0)._missilesLeft * 3 ))       //pos
//            );




}






void
IPlayableState::onHold()
{


}




void
IPlayableState::updatePlayer(const TPlayerId& playerId,const TTimeUnit& elapsedTime )
{



    //! set to true to refresh HUD rtt ?
//    getPlayer(playerId);
    TPlayerPair& player = getPlayer(playerId);
    IWalkyrieMode* mode = player.getWalkyrie()->getCurrentMode();


    auto _cachedInput = engine().getCachedInput();

    if( !player.getWalkyrie()->isTransforming() )
    {
        player.processTransformations( _cachedInput[ NPlayerInput::ModeUp ], _cachedInput[ NPlayerInput::ModeDown ]);
    }



//

    //////////////////////////////////////////////////
    /////   Update Position
    //////////////////////////////////////////////////
    // pos/rot/ mode
    TVector3df pos = player.getPosition();
    TVector3df rot = player.getRotation();

    _playerPreviousPos = pos;

    #define DECLARE_BIND(name) const SBindCache& name = _cachedInput[ NPlayerInput::name]


    DECLARE_BIND(MoveLeft);
    DECLARE_BIND(MoveRight);
    DECLARE_BIND(MoveForward);
    DECLARE_BIND(MoveBackward);


//    _INFO << "before animator call" << pos;
    if(player.UpdatePositionFunction != nullptr)
    {
//        bool ret =
        auto ret  = player.UpdatePositionFunction->processMove( PLAYER_UPDATE_FUNCTION_PARAMS );
        if ( ret != player.UpdatePositionFunction )
        {
            player.setAnimator(ret);
        }
    }
    else {
        player.setAnimator( mode->getDefaultAnimator() );
    }


//    _INFO << "after animator call" << pos;

    //
    clampPositionToAccessibleArea( pos );



//    _INFO << "after clamp" << pos;
    player.update(elapsedTime,_cachedInput);

    //
    player.setPosition( pos );
    player.setRotation( rot );

    //!

}


void
IPlayableState::onBatteryActivation(const IStaticBattery* battery)
{
    _INFO << "a battery got created";
}



void
IPlayableState::onBatteryDeath(const IStaticBattery* battery)
{
    //! battery killed
    // TODO should create explosion
    _INFO << __FUNCTION__ << " called";
}

void
IPlayableState::onCollision(const IStaticBattery* battery1, const IStaticBattery* battery2)
{
    _INFO << __FUNCTION__ << " called";
}



bool
IPlayableState::isEntityOufOfScreen(const IStaticBattery* battery) const
{
    const TVector3df& pos = battery->getPosition();
    const float Shift = 10;

    static core::rectf clippingPlane( core::vector2df(conf::_X_MIN, conf::_Z_MIN), core::vector2df(conf::_X_MAX, conf::_Z_MAX) );

//    return clippingPlane.isPointInside()
//    // TODO Ne pas le faire  faire a ttes les frames
    if( (pos.X > conf::_X_MAX ) || (pos.X < conf::_X_MIN) )
    {
        return true;
    }

    if( (pos.Z > conf::_Z_MAX + Shift) || (pos.Z < conf::_Z_MIN) )
    {
        return true;
    }

    return false;
}




void
IPlayableState::clampPositionToAccessibleArea(irr::core::vector3df& pos)
{
    pos.X = core::clamp(pos.X, conf::_X_MIN, conf::_X_MAX);
    pos.Z = core::clamp(pos.Z, conf::_Z_MIN, conf::_Z_MAX);

}



//void
//enableMenu( irr::gui::IGUITab* tab, const bool& enable)
//{
//    //
////    _INFO << "Enable menu [" << tab << "] or not: [" << enable << "]";
//    tab->setVisible(enable);
//    //tab->setEnabled(enable);
//    //engine().showCursor(enable);
//}


/// \function HandleEvents
bool
IPlayableState::handleMainEvents(const irr::SEvent& event)
{


    //! According to event type
    switch(event.EventType)
    {

        // Handle keys
        case EET_KEY_INPUT_EVENT :

            // check that's no key up
            if(!event.KeyInput.PressedDown)
                return false;

            switch (event.KeyInput.Key)
            {



                case irr::KEY_ESCAPE:
                    _INFO << "push pause menu";
                    engine().pushState( new CIngamePauseState() );
                    return true;

                // Cree une explosion
                // sous forme de sprite
                // Activate editor
//                case irr::KEY_KEY_E:
//
//                    _INFO << "activate editor";
//                    _INFO << "add sprite explosion (disabled)";
                    //_levelMgr.addSpriteExplosion(irr::core::vector3df(5,0,0));

                    break;


                // TODO afficher un shader de blur
                case KEY_KEY_B:
                    _INFO << "TODO: activer shader";
                    break;

                case KEY_F2:

                    _INFO << "Switch freeflight camera";
//                    // si on vient d'activer le mode free camera
//                    if( engine().SwitchBit(NBit::FreeFlight)){
//
//                        // pauses game
//                        //SwitchBit(NBit::Pause,true);
////                        scene::ICameraSceneNode* cam = smgr->setCamera();
////                        _cameras.push(cam);
//                        scene()->setActiveCamera( engine()._fpsCamera  );
//
//                    }
//                    else
//                    {
//                        scene()->setActiveCamera(_camera  );
//                    }
                    break;

                // cree une explosion avec spark
//                case irr::KEY_KEY_G:
//                    {
//                    //boost::any pos;
//                    //engine().addMessage(NMessageType::CreateExplosion,pos);
//
//                    //_levelMgr.addExplosion(irr::core::vector3df(5,0,0));
//                        _INFO << "Add spark explosion";
//                    //engine()._particleManager
//                        //ISceneNode* node = CGameEngine::getParticleManager()->createExplosion();
//                        //node->setParent( scene()->getRootSceneNode() );
//                        addSimpleExplosion( TVector3df(0,0,0) );
//                        this->shakeCamera();
//                    }
//                    break;
//
//                case irr::KEY_KEY_N:
//
//                    shakeCamera();
//                    break;

                case irr::KEY_F3:

                    ShowDebugData = !ShowDebugData;
                    _INFO << "Set show debug data to " << ShowDebugData;
                    break;

                // Tue le joueur
                case irr::KEY_KEY_M:

                    _INFO << "Mort du joueur";
                    getPlayer(0).kill();
//                    getLevelManager().updateState();
                    break;

                // Reload level
                case irr::KEY_KEY_R :

                    if(event.KeyInput.PressedDown)
                    {
                        // TODO reload script de base

//                        _INFO << "Reloading level...";
//                        _initialized = (setupLuaThread() && reloadLevel() );
//                        if(_initialized){
//                            _INFO << "Reset successful";
//                        }
//                        else {
//                            _LOG_WARNING << "Reseting context failed";
//                        }
                    }
                    break;

                // Reload
                case irr::KEY_KEY_D :
                // TODO tuer le joueur
                    //_player.Kill();
                    //_INFO << "setPlay state to dead";
                    //getLevelManager()._playState = CLevelManager::PlayerDead;

                    break;

                default:
                    break;
            }
            break;

        default:
            break;
    } // end 1nd switch



    //! Updater le player
//    _player->updateInput();


    return false;
}



